/*
    Copyright (C) 1999 by Joe Pasqua <joe@NoRoomAtTheInn.org>.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.


    THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
    OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
    OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

// NOTES:
//
// TO DO:
// * Add a method (asBasicVal) that returns the value as a BasicVal. Right now
//   some code is using toString() for this functionality. Migrate those
//   callers to this new method.

import java.util.Vector;

    /** This class represents an instance of a given type. Each actual value
        must subclass this class and must be named typeName + "Val". That is,
        the sublcass representing a Value named Number must be called NumberVal.
        That's so instances can be created dynamically by name without requiring
        a registration.
    */

class Value {

// ----------------------------------------------------------------------
// Class Methods
// ----------------------------------------------------------------------

    /** Parse a string and divide it into two parts divided by a separator.
        Each part will be a positive integer. If a character is found in
        the first part, it is converted to an int by subtracting 'A' or 'a'.
        The separator is a single character and can be any char found in the
        sep String. The parts must be in the ranges specified by the min/max
        pairs. The return value is an array of three elements. The 1st element
        is the value of the first part, the 2nd element is the value of the
        second part, and the 3rd element is the index into the sep String of
        the char that actually was found as the separator.
<BR>
        This method expects that the input String will begin and end with
        double quotes (").
    */
    static int[] parts(
            String whole, String sep,
            int p1Min, int p1Max, int p2Min, int p2Max)
            throws ParseException
    {
        int[] parts = new int[3];

        if (whole == null)
            throw new ParseException("Empty value");
        whole = whole.trim();
        if (whole.charAt(0) != '\"' || whole.charAt(whole.length()-1) != '\"')
            throw new ParseException("Malformed String Literal");
        int index = -1;
        for (int i = sep.length()-1; i >= 0; i--) {
            index = whole.indexOf(sep.charAt(i));
            if (index != -1)
                break;
        }
        if (index == -1 || index == 0 || index == whole.length()-1)
            throw new ParseException("Bad or missing Seperator");
        parts[2] = sep.indexOf(whole.charAt(index));
        String  p1 = whole.substring(1, index);
        String  p2 = whole.substring(index+1, whole.length()-1);
        parts[0] = -1;
        if (p1.length() == 1) {
            char c = p1.charAt(0);
            if (c >= 'A' && c <= 'Z')
                parts[0] = c - 'A';
            else if (c >= 'a' && c <= 'z')
                parts[0] = c - 'a';
        }
        try {
            if (parts[0] == -1)
                parts[0] = (new Integer(p1)).intValue();
            parts[1] = (new Integer(p2)).intValue();
            if (parts[0] < p1Min || parts[0] > p1Max ||
                parts[1] < p2Min || parts[1] > p2Max)
                    throw new ParseException("Value out of range");
        } catch (NumberFormatException e) {
            throw new ParseException("Bad Number Format");
        }

        return parts;
    }

// ----------------------------------------------------------------------
// Protected Instance Variables
// ----------------------------------------------------------------------

    /** The type of this instance                                       */
    protected Type  type;


// ----------------------------------------------------------------------
// Instance Methods which must be overriden
// ----------------------------------------------------------------------

    /** Initialize an instance with a given value. Subclasses will override
        this method to actually handle the value.
    */
    void init(Type t, String v)
            throws ParseException
    {
        this.type = t;
    }

    /** If the values of this type are orderable, then a subclass should
        override this method to implement a less than comparison
    */
    boolean lt(Value v) { return false; }

    /** If the values of this type are orderable, then a subclass should
        override this method to implement a greater than comparison
    */
    boolean gt(Value v) { return true; }


// ----------------------------------------------------------------------
// Instance Methods
// ----------------------------------------------------------------------

    /** Return the Type of this instance
    */
    Type getType() { return type; }



}
