/*
    Copyright (C) 1999 by Joe Pasqua <joe@NoRoomAtTheInn.org>.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.


    THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
    OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
    OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

import java.util.*;
import java.io.FileOutputStream;
import java.io.IOException;

    /** This class contains the main program for the XA Compiler.
        It is responsible for invoking the parser and then calling
        routines to generate the code.
    */

class XACompiler {
    static SymTab           symTab = new SymTab();
    static String           srcFile;
    static FileOutputStream fos;
    static int              bias;
    static Hashtable        userMethods = new Hashtable(5);

    public static void main(String args[]) {
        XAParser parser;

        if (args.length == 0) {
            usage();
            return;
        }

        if (args.length <= 2) {
            try {
                srcFile = args[0];
                parser = new XAParser(new java.io.FileInputStream(srcFile));
            } catch (java.io.FileNotFoundException e) {
                System.out.println("XACompiler Version 0.1:  File "
                        + args[0] + " not found.");
                return;
            }
        }
        else {
            usage();
            return;
        }

        try {
            parser.disable_tracing();
            if (args.length > 1 && args[1].equalsIgnoreCase("-trace"))
                parser.enable_tracing();

            predefine();


            Vector statements = parser.CompilationUnit();
            String  outfile = args[0];
            int     index = outfile.lastIndexOf('.');
            if (index == -1)
                outfile += ".pgm";
            else
                outfile = outfile.substring(0, index) + ".pgm";

            try {
                fos = new FileOutputStream(outfile);
            } catch (IOException e) {
                System.out.println("Couldn't access output file: " + e);
                return;
            }

            //dump(statements);
            genProgram(statements, fos);
        } catch (ParseException e) {
            System.out.println(e.getMessage());
            //e.printStackTrace();
        }
    }

    private static void usage() {
        System.out.println("usage: java XACompiler inputfile [-trace]");
    }

    static int mapOp(String op) {
        if (op.equals("if"))
            return 0;
        if (op.equals("&&"))
            return 0x0C;
        if (op.equals("||"))
            return 0x10;
        throw new RuntimeException("Unexpected conjunction: " + op);
    }

    static void genProgram(Vector runStatements, FileOutputStream fos)
            throws ParseException
    {
        int     lines = 0;
        short[] code;
        String  comment;

        IfStatement initStatement = initVars();
        if (initStatement != null)
            runStatements.insertElementAt(initStatement, 0);

        for (int i = 0; i < runStatements.size(); i++) {
            IfStatement conditional = (IfStatement)runStatements.elementAt(i);
            Vector terms = conditional.predicate.terms;
            Vector ops = conditional.predicate.ops;

            for (int j = 0; j < terms.size(); j++) {
                bias = mapOp((String)ops.elementAt(j));
                code = codeGen(bias, (MethodCall)terms.elementAt(j));
                if (code == null)
                    throw new RuntimeException(
                            "Got no code back for a conditional. " +
                            "This should only happen for user defined " +
                            "methods which can't be used in conditionals");
                comment = genComment(
                    (String)ops.elementAt(j), (MethodCall)terms.elementAt(j));
                emit(code, comment, fos);
                lines++;
            }

            bias = 0;
            Vector thens = conditional.thens;
            for (int j = 0; j < thens.size(); j++) {
                code = codeGen(0, (MethodCall)thens.elementAt(j));
                if (code != null) {
                    comment = genComment("Then", (MethodCall)thens.elementAt(j));
                    emit(code, comment, fos);
                }
                lines++;
            }

            bias = 4;
            Vector elses = conditional.elses;
            for (int j = 0; j < elses.size(); j++) {
                code = codeGen(bias, (MethodCall)elses.elementAt(j));
                if (code != null) {
                    comment = genComment("Else", (MethodCall)elses.elementAt(j));
                    emit(code, comment, fos);
                }
                lines++;
            }
        }

        short[] end = { 0xE0, 0x7F, 0, 0, 0, 0};
        short[] empty = { 0, 0, 0, 0, 0, 0 };

        emit(end, "END", fos); lines++;
        while (lines < 2049) {
            emit(empty, "(unused)", fos);
            lines++;
        }
    }

    static String genComment(String prefix, MethodCall m) {
        StringBuffer buf = new StringBuffer(80);
        buf.append(prefix);
        buf.append(' ');
        buf.append(m.toString());
        return buf.toString();
    }

    static void dump(Vector runStatements)
            throws ParseException
    {
        /*
        Enumeration e;
        e = symTab.keys();
        while (e.hasMoreElements())
            System.out.println("final " + symTab.get(e.nextElement()));

        e = symTab.keys();
        while (e.hasMoreElements())
            System.out.println(symTab.get(e.nextElement()));
        */

        System.out.println("\nrun {");
        for (int i = 0; i < runStatements.size(); i++) {
            IfStatement conditional = (IfStatement)runStatements.elementAt(i);
            System.out.println(conditional);
            Vector terms = conditional.predicate.terms;
            Vector ops = conditional.predicate.ops;
            short[] code;

            for (int j = 0; j < terms.size(); j++) {
                int bias = mapOp((String)ops.elementAt(j));
                code = codeGen(bias, (MethodCall)terms.elementAt(j));
                dumpCode(code);
            }

            Vector thens = conditional.thens;
            bias = 0;
            for (int j = 0; j < thens.size(); j++) {
                code = codeGen(bias, (MethodCall)thens.elementAt(j));
                dumpCode(code);
            }

            Vector elses = conditional.elses;
            bias = 8;
            for (int j = 0; j < elses.size(); j++) {
                code = codeGen(bias, (MethodCall)elses.elementAt(j));
                dumpCode(code);
            }

        }
        System.out.println("}");
    }

    static short[] codeGen(int opCodeBias, MethodCall method)
            throws ParseException
    {
        try {
            Symbol var = symTab.get(method.object);
            if (var == null)
                    throw new ParseException(
                            "Undefined variable: " + method.object);
            short[] code = var.type.codeGen(
                    var.initVal, method.method, method.args);
            if (code != null)
                code[1] += opCodeBias;
            return code;
        } catch (ParseException e) {
            throw new MethodCall.Error(e.getMessage(), method.srcLine);
        }
    }

    static void emit(short[] code, String comment, FileOutputStream fos) {
        try {
            int i;
            for (i = 0; i < 6; i++)
                fos.write(code[i]);

            int cLen = comment.length();
            if (cLen > 79) cLen = 79;
            for (i = 0; i < cLen; i++)
                fos.write(comment.charAt(i));
            for (; i < 79; i++)
                fos.write(0x20);
            fos.write(0x00);

        } catch (IOException e) {
            System.out.println("Failed writing to output file: " + e);
        }
    }

    static String toHex(int code) {
        return
            ((code < 16) ? "0x0" : "0x") +
            Integer.toHexString(code).toUpperCase();
    }

    static void dumpCode(short[] code) {
            System.out.println(
                    "\t\t[" + toHex(code[0]) + ", " +
                    toHex(code[1]) + "], [" +
                    toHex(code[2]) + ", " +
                    toHex(code[3]) + "], [" +
                    toHex(code[4]) + ", " +
                    toHex(code[5]) + "]");
    }

    static IfStatement initVars()
        throws ParseException
    {
        Vector  inits = new Vector(1);
        Vector  args;

        Enumeration e = symTab.keys();
        while (e.hasMoreElements()) {
            Symbol  s = symTab.get((String)e.nextElement());
            if (!s.isVar)
                continue;
            if (s.type.equals(Type.get("Number"))) {
                NumberVal   nv = (NumberVal)s.initVal;
                if (nv.intValue() != 0)
                    inits.addElement(s);
            }
            else if (s.type.equals(Type.get("Timer"))) {
                TimerVal    tv = (TimerVal)s.initVal;
                if (tv.timerVal() != 0)
                    inits.addElement(s);
            }
        }
        if (inits.size() == 0)
            return null;

        for (int i = inits.size()-1; i >= 0; i--) {
            args = new Vector(1);
            Symbol.Variable v = (Symbol.Variable)inits.elementAt(i);
            args.addElement(new BasicVal("="));
            args.addElement(new BasicVal(v.initVal.toString()));
            inits.setElementAt(new MethodCall(v.name, "set", args, 0), i);
        }
        args = new Vector(2);
        args.addElement(new BasicVal("="));
        args.addElement(new BasicVal("65535"));
        inits.addElement(new MethodCall("__init", "set", args, 0));

        args = new Vector(1);
        args.addElement(new BasicVal("=="));
        args.addElement(new BasicVal("0"));
        MethodCall  compare = new MethodCall("__init", "compare", args, 0);

        return new IfStatement(
                new BoolExpr(compare, "if"), inits, new Vector(1));
    }

    private static void predefine()
            throws ParseException
    {
        // Predefine some constants
        symTab.addConst("IsOff", "Number", "0");
        symTab.addConst("IsOn", "Number", "1");
        symTab.addConst("TurnsOff", "Number", "2");
        symTab.addConst("TurnsOn", "Number", "3");
        symTab.addConst("OnCmdPair", "Number", "4");
        symTab.addConst("OffCmdPair", "Number", "5");
        symTab.addConst("Unset", "Number", "0");
        symTab.addConst("True", "Number", "1");
        symTab.addConst("False", "Number", "2");
        symTab.addConst("On", "Number", "1");
        symTab.addConst("Off", "Number", "2");
        symTab.addConst("January", "Number", "1");
        symTab.addConst("February", "Number", "2");
        symTab.addConst("March", "Number", "3");
        symTab.addConst("April", "Number", "4");
        symTab.addConst("May", "Number", "5");
        symTab.addConst("June", "Number", "6");
        symTab.addConst("July", "Number", "7");
        symTab.addConst("August", "Number", "8");
        symTab.addConst("September", "Number", "9");
        symTab.addConst("October", "Number", "10");
        symTab.addConst("November", "Number", "11");
        symTab.addConst("December", "Number", "12");

        symTab.addConst("Sunday", "Number", "0");
        symTab.addConst("Monday", "Number", "1");
        symTab.addConst("Tuesday", "Number", "2");
        symTab.addConst("Wednesday", "Number", "3");
        symTab.addConst("Thursday", "Number", "4");
        symTab.addConst("Friday", "Number", "5");
        symTab.addConst("Saturday", "Number", "6");


        // Predefine some variables
        symTab.addVar("this", "XAControl");
        symTab.addVar("__init", "Number");
        symTab.addVar("cpuxa", "CPUXA");
        symTab.addVar("time", "Time");
        symTab.addVar("month", "Month");
        symTab.addVar("dayOfMonth", "DOM");
        symTab.addVar("dayOfWeek", "DOW");
        symTab.addVar("year", "Year");
        symTab.addVar("date", "Date");
    }

}